import json
import logging

from django.core.management import call_command
from django.db import connections
from django.test import Client, TransactionTestCase

from django_omnitenant.conf import settings
from django_omnitenant.models import BaseTenant
from django_omnitenant.tenant_context import TenantContext
from django_omnitenant.utils import get_domain_model, get_tenant_model

logger = logging.getLogger(__name__)


class TenantAPITestCaseMixin:
    client_class = Client

    def _request(
        self,
        method: str,
        path: str,
        data=None,
        headers=None,
        content_type: str = "application/json",
        **kwargs,
    ):
        headers = headers or {}
        if method.lower() == "get":
            return self.client.get(path, data or {}, **headers, **kwargs)
        body = json.dumps(data or {}) if content_type == "application/json" else data
        return getattr(self.client, method)(
            path,
            body,
            content_type=content_type,
            **headers,
            **kwargs,
        )

    def get(self, path, data=None, headers=None, **kwargs):
        return self._request("get", path, data, headers, **kwargs)

    def post(
        self, path, data=None, headers=None, content_type="application/json", **kwargs
    ):
        return self._request("post", path, data, headers, content_type, **kwargs)

    def put(
        self, path, data=None, headers=None, content_type="application/json", **kwargs
    ):
        return self._request("put", path, data, headers, content_type, **kwargs)

    def patch(
        self, path, data=None, headers=None, content_type="application/json", **kwargs
    ):
        return self._request("patch", path, data, headers, content_type, **kwargs)

    def delete(self, path, data=None, headers=None, **kwargs):
        return self._request("delete", path, data, headers, **kwargs)


class BaseTenantTestCase(TransactionTestCase):
    tenant = None
    keepdb = True
    flush_per_test = False
    flush_per_class = True
    databases = {"default"}

    @classmethod
    def _setup_tenant(cls):
        raise NotImplementedError

    @classmethod
    def _remove_tenant(cls):
        if not cls.tenant:
            return
        try:
            cls.tenant.delete()
        except Exception as e:
            print(
                f"Error deleting tenant object '{cls.tenant}': {e}"
            )  # Use print for simplicity
        finally:
            cls.tenant = None

    @classmethod
    def setUpClass(cls) -> None:
        cls._setup_tenant()
        cls._tenant_context = TenantContext.use_tenant(cls.tenant)
        cls._tenant_context.__enter__()
        super().setUpClass()

    @classmethod
    def tearDownClass(cls) -> None:
        tenant_ctx = getattr(cls, "_tenant_context", None)
        if tenant_ctx is not None:
            try:
                tenant_ctx.__exit__(None, None, None)
            except Exception as e:
                print(f"Error exiting tenant context: {e}")
            finally:
                cls._tenant_context = None
        cls._remove_tenant()
        super().tearDownClass()


class BaseAPITestCase(TenantAPITestCaseMixin, BaseTenantTestCase):
    domain = None

    @classmethod
    def _setup_domain(cls):
        if cls.domain is None:
            Domain = get_domain_model()
            cls.domain = Domain.objects.create(
                tenant=cls.tenant,
                domain=settings.DEFAULT_HOST,
            )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls._setup_domain()

    def setUp(self):
        super().setUp()
        self.client = self.client_class()


class DBTenantTestCase(BaseTenantTestCase):
    databases = {"default", "test_tenant"}  # Ensure non-default alias
    _db_alias = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if cls.flush_per_class and cls._db_alias:
            cls._flush_db(cls._db_alias)

    @classmethod
    def tearDownClass(cls):
        if not cls.keepdb and cls._db_alias:
            cls._destroy_db()
        super().tearDownClass()

    def tearDown(self):
        if self.keepdb and self.flush_per_test and self.__class__._db_alias:
            self._flush_db(self.__class__._db_alias)
        super().tearDown()

    @classmethod
    def _setup_tenant(cls):
        if cls.tenant is None:
            Tenant = get_tenant_model()
            cls.tenant = Tenant.objects.create(
                name="Test Tenant",
                tenant_id="test_db_tenant",
                isolation_type=BaseTenant.IsolationType.DATABASE,
            )

        cls._db_alias = cls._get_db_alias()
        if cls._db_alias not in cls.databases:
            cls.databases = set(cls.databases) | {cls._db_alias}
        conn = connections[cls._db_alias]
        conn.creation.create_test_db(
            autoclobber=True, serialize=False, keepdb=cls.keepdb
        )

    @classmethod
    def _destroy_db(cls):
        conn = connections[cls._db_alias]
        try:
            conn.close()
            conn.creation.destroy_test_db(cls._db_alias, verbosity=0)
        finally:
            cls._db_alias = None
            cls.databases = {"default"}

    @classmethod
    def _get_db_alias(cls):
        db_alias = next((db for db in cls.databases if db != "default"), "test_tenant")
        config = cls.tenant.config.get("db_config", {})
        config["NAME"] = db_alias
        cls.tenant.config["db_config"] = config
        cls.tenant.save()
        return db_alias

    @staticmethod
    def _flush_db(db_alias):
        call_command("flush", verbosity=0, interactive=False, database=db_alias)


class DBTenantAPITestCase(BaseAPITestCase, DBTenantTestCase):
    pass


class SchemaTenantTestCase(BaseTenantTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        if cls.flush_per_class:
            cls._flush_schema()
        with TenantContext.use_tenant(cls.tenant):  # Run migrations in tenant context
            call_command("migrate", verbosity=0, interactive=False)

    def setUp(self) -> None:
        super().setUp()
        self._tenant_context = TenantContext.use_tenant(self.tenant)
        self._tenant_context.__enter__()

    def tearDown(self):
        if hasattr(self, "_tenant_context") and self._tenant_context is not None:
            self._tenant_context.__exit__(None, None, None)
            self._tenant_context = None
        super().tearDown()

    @classmethod
    def _setup_tenant(cls):
        if cls.tenant is None:
            Tenant = get_tenant_model()
            cls.tenant = Tenant.objects.create(
                name="Test Schema Tenant",
                tenant_id="test_schema_tenant",
                isolation_type=BaseTenant.IsolationType.SCHEMA,
            )

    @classmethod
    def _flush_schema(cls):
        with TenantContext.use_tenant(cls.tenant):
            call_command("flush", verbosity=0, interactive=False, database="default")


class SchemaTenantAPITestCase(BaseAPITestCase, SchemaTenantTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        with TenantContext.use_schema(settings.DEFAULT_TENANT_NAME):
            cls._setup_domain()



import re
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _
PGSQL_VALID_SCHEMA_NAME = re.compile(r"^(?!pg_).{1,63}$", re.IGNORECASE)


def is_valid_schema_name(name):
    return PGSQL_VALID_SCHEMA_NAME.match(name)


def _check_schema_name(name):
    if not is_valid_schema_name(name):
        raise ValidationError("Invalid string used for the schema name.")




def validate_dns_label(value):
    """
    Validate a single DNS label according to RFC 1034/1035:
    - Only letters, digits, and hyphens.
    - Cannot start or end with a hyphen.
    - Length between 1 and 63 characters.
    """
    if not re.match(r"^(?!-)[A-Za-z0-9-]{1,63}(?<!-)$", value):
        raise ValidationError(
            _("%(value)s is not a valid DNS label."),
            params={"value": value},
        )

def validate_domain_name(value):
    """
    Validate a full domain name (FQDN).
    - Split into labels and validate each one.
    - Total length must not exceed 253 characters.
    """
    if len(value) > 253:
        raise ValidationError(
            _("%(value)s exceeds the maximum length of 253 characters."),
            params={"value": value},
        )

    labels = value.split(".")

    for label in labels:
        validate_dns_label(label)



from contextlib import contextmanager
from contextvars import ContextVar

from django_omnitenant.conf import settings
from django_omnitenant.constants import constants
from django_omnitenant.models import BaseTenant
from django_omnitenant.utils import get_tenant_model


class TenantContext:
    _tenant_stack = ContextVar("tenant_stack", default=[])
    _db_alias_stack = ContextVar("db_alias_stack", default=[constants.DEFAULT_DB_ALIAS])
    _cache_alias_stack = ContextVar(
        "cache_alias_stack", default=[constants.DEFAULT_CACHE_ALIAS]
    )

    # --- Tenant ---
    @classmethod
    def get_tenant(cls) -> BaseTenant | None:
        stack = cls._tenant_stack.get()
        return stack[-1] if stack else None

    @classmethod
    def push_tenant(cls, tenant: BaseTenant):
        stack = cls._tenant_stack.get()
        new_stack = stack + [tenant]
        cls._tenant_stack.set(new_stack)

    @classmethod
    def pop_tenant(cls):
        stack = cls._tenant_stack.get()
        if stack:
            new_stack = stack[:-1]
            cls._tenant_stack.set(new_stack)

    # --- Database ---
    @classmethod
    def get_db_alias(cls):
        stack = cls._db_alias_stack.get()
        return stack[-1] if stack else constants.DEFAULT_DB_ALIAS

    @classmethod
    def push_db_alias(cls, db_alias):
        stack = cls._db_alias_stack.get()
        new_stack = stack + [db_alias]
        cls._db_alias_stack.set(new_stack)

    @classmethod
    def pop_db_alias(cls):
        stack = cls._db_alias_stack.get()
        if stack:
            new_stack = stack[:-1]
            cls._db_alias_stack.set(new_stack)

    # --- Cache ---
    @classmethod
    def get_cache_alias(cls):
        stack = cls._cache_alias_stack.get()
        return stack[-1] if stack else "default"

    @classmethod
    def push_cache_alias(cls, cache_alias):
        stack = cls._cache_alias_stack.get()
        new_stack = stack + [cache_alias]
        cls._cache_alias_stack.set(new_stack)

    @classmethod
    def pop_cache_alias(cls):
        stack = cls._cache_alias_stack.get()
        if stack:
            new_stack = stack[:-1]
            cls._cache_alias_stack.set(new_stack)

    # --- Clear all (reset to defaults) ---
    @classmethod
    def clear_all(cls):
        cls._tenant_stack.set([])
        cls._db_alias_stack.set([constants.DEFAULT_DB_ALIAS])
        cls._cache_alias_stack.set(["default"])

    # --- Context manager ---
    @classmethod
    @contextmanager
    def use_tenant(cls, tenant):
        from django_omnitenant.backends.cache_backend import CacheTenantBackend
        from django_omnitenant.backends.database_backend import DatabaseTenantBackend
        from django_omnitenant.backends.schema_backend import SchemaTenantBackend

        # Push tenant
        cls.push_tenant(tenant)

        # Activate DB/Schema backend
        backend = (
            SchemaTenantBackend(tenant)
            if tenant.isolation_type == BaseTenant.IsolationType.SCHEMA
            else DatabaseTenantBackend(tenant)
        )
        backend.activate()
        cls.push_db_alias(cls.get_db_alias())  # backend may change alias

        # Activate cache backend
        cache_backend = CacheTenantBackend(tenant)
        cache_backend.activate()
        cls.push_cache_alias(cls.get_cache_alias())

        try:
            yield
        finally:
            # Deactivate backends
            backend.deactivate()
            cache_backend.deactivate()

            # Pop tenant/db/cache
            cls.pop_tenant()
            cls.pop_db_alias()
            cls.pop_cache_alias()

    @classmethod
    @contextmanager
    def use_schema(cls, schema_name: str):
        """
        Context manager to use a specific schema.
        """
        from django_omnitenant.backends.schema_backend import SchemaTenantBackend

        tenant: BaseTenant = get_tenant_model()(tenant_id=schema_name)  # type: ignore # Mock tenant for context
        backend = SchemaTenantBackend(tenant)
        backend.activate()

        try:
            yield
        finally:
            backend.deactivate()
            cls.pop_db_alias()

    @classmethod
    @contextmanager
    def use_default_db(cls):
        from django_omnitenant.backends.cache_backend import CacheTenantBackend
        from django_omnitenant.backends.database_backend import DatabaseTenantBackend

        # Push default DB & cache
        default_db = constants.DEFAULT_DB_ALIAS
        default_cache = constants.DEFAULT_CACHE_ALIAS

        cls.push_db_alias(default_db)
        cls.push_cache_alias(default_cache)

        # Activate default backends
        tenant: BaseTenant = get_tenant_model()(tenant_id=settings.DEFAULT_TENANT_NAME)
        db_backend = DatabaseTenantBackend(tenant)  # None means no specific tenant
        db_backend.activate()
        cache_backend = CacheTenantBackend(tenant)
        cache_backend.activate()

        try:
            yield
        finally:
            db_backend.deactivate()
            cache_backend.deactivate()
            cls.pop_db_alias()
            cls.pop_cache_alias()

    # --- New: use public schema ---
    @classmethod
    @contextmanager
    def use_public_schema(cls):
        from django_omnitenant.backends.cache_backend import CacheTenantBackend
        from django_omnitenant.backends.schema_backend import SchemaTenantBackend

        # Create a mock tenant representing public schema
        tenant: BaseTenant = get_tenant_model()(tenant_id="public")  # type: ignore
        backend = SchemaTenantBackend(tenant)
        backend.activate()
        cls.push_db_alias(cls.get_db_alias())

        # Public cache
        cache_backend = CacheTenantBackend(tenant)
        cache_backend.activate()
        cls.push_cache_alias(cls.get_cache_alias())

        try:
            yield
        finally:
            backend.deactivate()
            cache_backend.deactivate()
            cls.pop_db_alias()
            cls.pop_cache_alias()
